"""System Bridge: GUI"""
import asyncio
from logging import Logger
import subprocess
import sys
from threading import Thread

from systembridgeshared.exceptions import ConnectionErrorException
from systembridgeshared.settings import Settings
from systembridgeshared.websocket_client import WebSocketClient


class GUIAttemptsExceededException(BaseException):
    """Raise this when the GUI attempts to start more too many times."""


async def start_gui(
    logger: Logger,
    settings: Settings,
    attempt=1,
) -> None:
    """Start the GUI"""
    if attempt > 2:
        logger.error("Failed to start GUI after 2 attempts")
        raise GUIAttemptsExceededException("Failed to start GUI after 3 attempts")
    logger.info("Test WebSocket connection before starting GUI. Attempt #%s", attempt)
    websocket_client = WebSocketClient(settings)
    try:
        await websocket_client.connect()
        await websocket_client.close()
    except ConnectionErrorException:
        logger.warning("Could not connect to WebSocket. Retrying in 5 seconds")
        await asyncio.sleep(5)
        await start_gui(logger, settings, attempt + 1)
        return

    logger.info("Executable: %s", sys.executable)
    logger.info("Starting GUI")
    with subprocess.Popen(
        [
            sys.executable,
            "-m",
            "systembridgegui",
        ]
    ) as process:
        logger.info("GUI started with PID: %s", process.pid)
        if (exit_code := process.wait()) != 0:
            logger.error("GUI exited with code: %s", exit_code)
            await start_gui(logger, settings, attempt + 1)
            return
        logger.info("GUI exited with code: %s", exit_code)


def start_gui_sync(
    logger: Logger,
    settings: Settings,
) -> None:
    """Start the GUI in a synchronous thread"""
    asyncio.run(start_gui(logger, settings))


async def start_gui_threaded(
    logger: Logger,
    settings: Settings,
) -> None:
    """Start the GUI in a thread"""
    thread = Thread(target=start_gui_sync, args=(logger, settings))
    thread.start()



"""System Bridge: Data"""
import asyncio
from collections.abc import Callable
from threading import Thread

from systembridgeshared.base import Base
from systembridgeshared.database import Database

from systembridgebackend.modules.update import Update


class UpdateThread(Thread):
    """Update thread"""

    def __init__(
        self,
        database: Database,
        updated_callback: Callable,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database
        self._update = Update(self._database)
        self._updated_callback = updated_callback

    def run(self) -> None:
        """Run"""
        asyncio.run(self._update.update_data(self._updated_callback))


class UpdateFrequentThread(Thread):
    """Update frequent thread"""

    def __init__(
        self,
        database: Database,
        updated_callback: Callable,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database
        self._update = Update(self._database)
        self._updated_callback = updated_callback

    def run(self) -> None:
        """Run"""
        asyncio.run(self._update.update_frequent_data(self._updated_callback))


class Data(Base):
    """Data"""

    def __init__(
        self,
        database: Database,
        updated_callback: Callable,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database
        self._updated_callback = updated_callback

    def request_update_data(self) -> None:
        """Request update data"""
        thread = UpdateThread(
            self._database,
            self._updated_callback,
        )
        thread.start()

    def request_update_frequent_data(self) -> None:
        """Request update frequent data"""
        thread = UpdateFrequentThread(
            self._database,
            self._updated_callback,
        )
        thread.start()



"""System Bridge: WebSocket handler"""
from collections.abc import Callable
from json import JSONDecodeError, dumps, loads
import os
from uuid import uuid4

from systembridgeshared.base import Base
from systembridgeshared.const import (
    EVENT_BASE,
    EVENT_DATA,
    EVENT_DIRECTORIES,
    EVENT_EVENT,
    EVENT_FILE,
    EVENT_FILES,
    EVENT_ID,
    EVENT_MESSAGE,
    EVENT_MODULE,
    EVENT_MODULES,
    EVENT_PATH,
    EVENT_SETTING,
    EVENT_SUBTYPE,
    EVENT_TYPE,
    EVENT_VALUE,
    EVENT_VERSIONS,
    SETTING_AUTOSTART,
    SUBTYPE_BAD_API_KEY,
    SUBTYPE_BAD_DIRECTORY,
    SUBTYPE_BAD_FILE,
    SUBTYPE_BAD_JSON,
    SUBTYPE_BAD_PATH,
    SUBTYPE_LISTENER_ALREADY_REGISTERED,
    SUBTYPE_LISTENER_NOT_REGISTERED,
    SUBTYPE_MISSING_API_KEY,
    SUBTYPE_MISSING_BASE,
    SUBTYPE_MISSING_KEY,
    SUBTYPE_MISSING_MODULES,
    SUBTYPE_MISSING_PATH,
    SUBTYPE_MISSING_PATH_URL,
    SUBTYPE_MISSING_SETTING,
    SUBTYPE_MISSING_TEXT,
    SUBTYPE_MISSING_VALUE,
    SUBTYPE_UNKNOWN_EVENT,
    TYPE_APPLICATION_UPDATE,
    TYPE_APPLICATION_UPDATING,
    TYPE_DATA_GET,
    TYPE_DATA_LISTENER_REGISTERED,
    TYPE_DATA_LISTENER_UNREGISTERED,
    TYPE_DATA_UPDATE,
    TYPE_DIRECTORIES,
    TYPE_ERROR,
    TYPE_EXIT_APPLICATION,
    TYPE_FILE,
    TYPE_FILES,
    TYPE_GET_DATA,
    TYPE_GET_DIRECTORIES,
    TYPE_GET_FILE,
    TYPE_GET_FILES,
    TYPE_GET_SETTING,
    TYPE_GET_SETTINGS,
    TYPE_KEYBOARD_KEY_PRESSED,
    TYPE_KEYBOARD_KEYPRESS,
    TYPE_KEYBOARD_TEXT,
    TYPE_KEYBOARD_TEXT_SENT,
    TYPE_OPEN,
    TYPE_OPENED,
    TYPE_POWER_HIBERNATE,
    TYPE_POWER_HIBERNATING,
    TYPE_POWER_LOCK,
    TYPE_POWER_LOCKING,
    TYPE_POWER_LOGGINGOUT,
    TYPE_POWER_LOGOUT,
    TYPE_POWER_RESTART,
    TYPE_POWER_RESTARTING,
    TYPE_POWER_SHUTDOWN,
    TYPE_POWER_SHUTTINGDOWN,
    TYPE_POWER_SLEEP,
    TYPE_POWER_SLEEPING,
    TYPE_REGISTER_DATA_LISTENER,
    TYPE_SETTING_RESULT,
    TYPE_SETTING_UPDATED,
    TYPE_SETTINGS_RESULT,
    TYPE_UNREGISTER_DATA_LISTENER,
    TYPE_UPDATE_SETTING,
)
from systembridgeshared.database import Database
from systembridgeshared.settings import SECRET_API_KEY, Settings
from systembridgeshared.update import Update

from systembridgebackend.autostart import autostart_disable, autostart_enable
from systembridgebackend.modules.listeners import Listeners
from systembridgebackend.server.keyboard import keyboard_keypress, keyboard_text
from systembridgebackend.server.media import (
    BASE_DIRECTORIES,
    get_directories,
    get_file,
    get_files,
)
from systembridgebackend.server.open import open_path, open_url
from systembridgebackend.server.power import (
    hibernate,
    lock,
    logout,
    restart,
    shutdown,
    sleep,
)


class WebSocketHandler(Base):
    """WebSocket handler"""

    def __init__(
        self,
        database: Database,
        settings: Settings,
        listeners: Listeners,
        implemented_modules: list[str],  # pylint: disable=unsubscriptable-object
        websocket,
        callback_exit_application: Callable,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database
        self._settings = settings
        self._listeners = listeners
        self._implemented_modules = implemented_modules
        self._websocket = websocket
        self._callback_exit_application = callback_exit_application
        self._active = True

    async def _check_api_key(
        self,
        data: dict,
    ) -> bool:
        """Check API key"""
        if "api-key" not in data:
            self._logger.warning("No api-key provided")
            await self._websocket.send(
                dumps(
                    {
                        EVENT_TYPE: TYPE_ERROR,
                        EVENT_SUBTYPE: SUBTYPE_MISSING_API_KEY,
                        EVENT_MESSAGE: "No api-key provided",
                    }
                )
            )
            return False
        if data["api-key"] != self._settings.get_secret(SECRET_API_KEY):
            self._logger.warning("Invalid api-key")
            await self._websocket.send(
                dumps(
                    {
                        EVENT_TYPE: TYPE_ERROR,
                        EVENT_SUBTYPE: SUBTYPE_BAD_API_KEY,
                        EVENT_MESSAGE: "Invalid api-key",
                    }
                )
            )
            return False
        return True

    async def _data_changed(
        self,
        module: str,
        data: dict,
    ) -> None:
        """Data changed"""
        if module not in self._implemented_modules:
            self._logger.info("Data module %s not in registered modules", module)
            return
        await self._websocket.send(
            dumps(
                {
                    EVENT_TYPE: TYPE_DATA_UPDATE,
                    EVENT_MESSAGE: "Data changed",
                    EVENT_MODULE: module,
                    EVENT_DATA: data,
                }
            )
        )

    async def _handler(
        self,
        listener_id: str,
    ) -> None:
        """Handler"""
        # Loop until the connection is closed
        while self._active:
            try:
                data = loads(await self._websocket.recv())
            except JSONDecodeError as error:
                self._logger.error("Invalid JSON: %s", error)
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_BAD_JSON,
                            EVENT_MESSAGE: "Invalid JSON",
                        }
                    )
                )
                continue

            self._logger.info("Received: %s", data[EVENT_EVENT])

            if not await self._check_api_key(data):
                continue

            if data[EVENT_EVENT] == TYPE_APPLICATION_UPDATE:
                versions = Update().update(
                    data.get("version"),
                    wait=False,
                )
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_APPLICATION_UPDATING,
                            EVENT_MESSAGE: "Updating application",
                            EVENT_VERSIONS: versions,
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_EXIT_APPLICATION:
                self._callback_exit_application()
                self._logger.info("Exit application called")
            elif data[EVENT_EVENT] == TYPE_KEYBOARD_KEYPRESS:
                if "key" not in data:
                    self._logger.warning("No key provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_KEY,
                                EVENT_MESSAGE: "No key provided",
                            }
                        )
                    )
                    continue

                try:
                    keyboard_keypress(data["key"])
                except ValueError as err:
                    self._logger.warning(err.args[0])
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_KEY,
                                EVENT_MESSAGE: "Invalid key",
                            }
                        )
                    )
                    continue

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_KEYBOARD_KEY_PRESSED,
                            EVENT_MESSAGE: "Key pressed",
                            EVENT_ID: listener_id,
                            "key": data["key"],
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_KEYBOARD_TEXT:
                if "text" not in data:
                    self._logger.warning("No text provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_TEXT,
                                EVENT_MESSAGE: "No text provided",
                            }
                        )
                    )
                    continue

                keyboard_text(data["text"])

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_KEYBOARD_TEXT_SENT,
                            EVENT_MESSAGE: "Key pressed",
                            EVENT_ID: listener_id,
                            "text": data["text"],
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_OPEN:
                if "path" in data:
                    open_path(data["path"])
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_OPENED,
                                EVENT_MESSAGE: "Path opened",
                                EVENT_ID: listener_id,
                                "path": data["path"],
                            }
                        )
                    )
                    continue
                if "url" in data:
                    open_url(data["url"])
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_OPENED,
                                EVENT_MESSAGE: "URL opened",
                                EVENT_ID: listener_id,
                                "url": data["url"],
                            }
                        )
                    )
                    continue

                self._logger.warning("No path or url provided")
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_MISSING_PATH_URL,
                            EVENT_MESSAGE: "No path or url provided",
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_REGISTER_DATA_LISTENER:
                if EVENT_MODULES not in data:
                    self._logger.warning("No modules provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_MODULES,
                                EVENT_MESSAGE: "No modules provided",
                            }
                        )
                    )
                    continue

                self._logger.info(
                    "Registering data listener: %s - %s",
                    listener_id,
                    data[EVENT_MODULES],
                )

                if await self._listeners.add_listener(
                    listener_id,
                    self._data_changed,
                    data[EVENT_MODULES],
                ):
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_LISTENER_ALREADY_REGISTERED,
                                EVENT_MESSAGE: "Listener already registered with this connection",
                                EVENT_ID: listener_id,
                                EVENT_MODULES: data[EVENT_MODULES],
                            }
                        )
                    )
                    continue

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_DATA_LISTENER_REGISTERED,
                            EVENT_MESSAGE: "Data listener registered",
                            EVENT_ID: listener_id,
                            EVENT_MODULES: data[EVENT_MODULES],
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_UNREGISTER_DATA_LISTENER:
                self._logger.info("Unregistering data listener %s", listener_id)

                if not self._listeners.remove_listener(listener_id):
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_LISTENER_NOT_REGISTERED,
                                EVENT_MESSAGE: "Listener not registered with this connection",
                            }
                        )
                    )
                    continue

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_DATA_LISTENER_UNREGISTERED,
                            EVENT_MESSAGE: "Data listener unregistered",
                            EVENT_ID: listener_id,
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_GET_DATA:
                if EVENT_MODULES not in data:
                    self._logger.warning("No modules provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_MODULES,
                                EVENT_MESSAGE: "No modules provided",
                            }
                        )
                    )
                    continue
                self._logger.info("Getting data: %s", data[EVENT_MODULES])

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_DATA_GET,
                            EVENT_MESSAGE: "Getting data",
                            EVENT_MODULES: data[EVENT_MODULES],
                        }
                    )
                )

                for module in data[EVENT_MODULES]:
                    data = self._database.table_data_to_ordered_dict(module)
                    if data is not None:
                        await self._websocket.send(
                            dumps(
                                {
                                    EVENT_TYPE: TYPE_DATA_UPDATE,
                                    EVENT_MESSAGE: "Data received",
                                    EVENT_MODULE: module,
                                    EVENT_DATA: data,
                                }
                            )
                        )
            elif data[EVENT_EVENT] == TYPE_GET_DIRECTORIES:
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_DIRECTORIES,
                            EVENT_DIRECTORIES: get_directories(),
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_GET_FILES:
                if EVENT_BASE not in data:
                    self._logger.warning("No base provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_BASE,
                                EVENT_MESSAGE: "No base provided",
                            }
                        )
                    )
                    continue
                path = (
                    os.path.join(BASE_DIRECTORIES[data[EVENT_BASE]], data[EVENT_PATH])
                    if data.get(EVENT_PATH)
                    else BASE_DIRECTORIES[data[EVENT_BASE]]
                )

                self._logger.info(
                    "Getting files: %s - %s - %s",
                    data[EVENT_BASE],
                    data.get(EVENT_PATH),
                    path,
                )

                if not os.path.exists(path):
                    self._logger.warning("Cannot find path")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_PATH,
                                EVENT_MESSAGE: "Cannot find path",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue
                if not os.path.isdir(path):
                    self._logger.warning("Path is not a directory")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_DIRECTORY,
                                EVENT_MESSAGE: "Path is not a directory",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_FILES,
                            EVENT_FILES: get_files(data[EVENT_BASE], path),
                            EVENT_PATH: path,
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_GET_FILE:
                if EVENT_BASE not in data:
                    self._logger.warning("No base provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_BASE,
                                EVENT_MESSAGE: "No base provided",
                            }
                        )
                    )
                    continue
                if EVENT_PATH not in data:
                    self._logger.warning("No path provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_PATH,
                                EVENT_MESSAGE: "No path provided",
                            }
                        )
                    )
                    continue
                path = os.path.join(
                    BASE_DIRECTORIES[data[EVENT_BASE]], data[EVENT_PATH]
                )

                self._logger.info(
                    "Getting file: %s - %s - %s",
                    data[EVENT_BASE],
                    data[EVENT_PATH],
                    path,
                )

                if not os.path.exists(path):
                    self._logger.warning("Cannot find path")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_PATH,
                                EVENT_MESSAGE: "Cannot find path",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue
                if not os.path.isfile(path):
                    self._logger.warning("Path is not a file")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_FILE,
                                EVENT_MESSAGE: "Path is not a file",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_FILE,
                            EVENT_FILE: get_file(
                                BASE_DIRECTORIES[data[EVENT_BASE]], path
                            ),
                            EVENT_PATH: path,
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_GET_SETTINGS:
                self._logger.info("Getting settings")

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_SETTINGS_RESULT,
                            EVENT_MESSAGE: "Got settings",
                            EVENT_DATA: self._settings.get_all(),
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_GET_SETTING:
                if EVENT_SETTING not in data:
                    self._logger.warning("No setting provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_SETTING,
                                EVENT_MESSAGE: "No setting provided",
                            }
                        )
                    )
                    continue
                self._logger.info("Getting setting: %s", data[EVENT_SETTING])

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_SETTING_RESULT,
                            EVENT_MESSAGE: "Got setting",
                            EVENT_SETTING: data[EVENT_SETTING],
                            EVENT_DATA: self._settings.get(data[EVENT_SETTING]),
                        }
                    )
                )
            elif data[EVENT_EVENT] == TYPE_UPDATE_SETTING:
                if EVENT_SETTING not in data:
                    self._logger.warning("No setting provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_SETTING,
                                EVENT_MESSAGE: "No setting provided",
                            }
                        )
                    )
                    continue
                if EVENT_VALUE not in data:
                    self._logger.warning("No value provided")
                    await self._websocket.send(
                        dumps(
                            {
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_VALUE,
                                EVENT_MESSAGE: "No value provided",
                            }
                        )
                    )
                    continue
                self._logger.info(
                    "Setting setting %s to: %s",
                    data[EVENT_SETTING],
                    data[EVENT_VALUE],
                )

                self._settings.set(data[EVENT_SETTING], data[EVENT_VALUE])

                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_SETTING_UPDATED,
                            EVENT_MESSAGE: "Setting updated",
                            EVENT_SETTING: data[EVENT_SETTING],
                            EVENT_VALUE: data[EVENT_VALUE],
                        }
                    )
                )

                if data[EVENT_SETTING] != SETTING_AUTOSTART:
                    continue
                self._logger.info("Setting autostart to %s", data[EVENT_VALUE])
                if data[EVENT_VALUE]:
                    autostart_enable()
                else:
                    autostart_disable()
            elif data[EVENT_EVENT] == TYPE_POWER_SLEEP:
                self._logger.info("Sleeping")
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_POWER_SLEEPING,
                            EVENT_MESSAGE: "Sleeping",
                        }
                    )
                )
                sleep()
            elif data[EVENT_EVENT] == TYPE_POWER_HIBERNATE:
                self._logger.info("Sleeping")
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_POWER_HIBERNATING,
                            EVENT_MESSAGE: "Hiibernating",
                        }
                    )
                )
                hibernate()
            elif data[EVENT_EVENT] == TYPE_POWER_RESTART:
                self._logger.info("Sleeping")
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_POWER_RESTARTING,
                            EVENT_MESSAGE: "Restarting",
                        }
                    )
                )
                restart()
            elif data[EVENT_EVENT] == TYPE_POWER_SHUTDOWN:
                self._logger.info("Sleeping")
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_POWER_SHUTTINGDOWN,
                            EVENT_MESSAGE: "Shutting down",
                        }
                    )
                )
                shutdown()
            elif data[EVENT_EVENT] == TYPE_POWER_LOCK:
                self._logger.info("Locking")
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_POWER_LOCKING,
                            EVENT_MESSAGE: "Locking",
                        }
                    )
                )
                lock()
            elif data[EVENT_EVENT] == TYPE_POWER_LOGOUT:
                self._logger.info("Logging out")
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_POWER_LOGGINGOUT,
                            EVENT_MESSAGE: "Logging out",
                        }
                    )
                )
                logout()
            else:
                self._logger.warning("Unknown event: %s", data[EVENT_EVENT])
                await self._websocket.send(
                    dumps(
                        {
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_UNKNOWN_EVENT,
                            EVENT_MESSAGE: "Unknown event",
                            EVENT_EVENT: data[EVENT_EVENT],
                        }
                    )
                )

    async def handler(self) -> None:
        """Handler"""
        listener_id = str(uuid4())
        try:
            await self._handler(listener_id)
        except ConnectionError as error:
            self._logger.info("Connection closed: %s", error)
        finally:
            self._logger.info("Unregistering data listener %s", listener_id)
            self._listeners.remove_listener(listener_id)

    def set_active(
        self,
        active: bool,
    ) -> None:
        """Set active"""
        self._active = active



"""System Bridge: Shortcut Linux"""
import os
import sys

desktop_entry = f"""
[Desktop Entry]
Name=System Bridge
Comment=System Bridge
Exec={sys.executable} -m systembridgebackend
Icon={os.path.join(os.path.dirname(__file__),'../icon.png')}
Terminal=false
Type=Application
Categories=Application;
"""


def create_linux_shortcuts():
    """Create Linux shortcuts"""
    path = os.path.expanduser("~/.local/share/applications/systembridge.desktop")
    if not os.path.exists(path):
        with open(path, "w", encoding="utf-8") as file:
            file.write(desktop_entry)



"""System Bridge: Update Battery"""
import asyncio

from systembridgeshared.common import camel_to_snake
from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.battery import Battery


class BatteryUpdate(ModuleUpdateBase):
    """Battery Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "battery")
        self._battery = Battery()

    async def update_sensors(self) -> None:
        """Update Battery Sensors"""
        if data := self._battery.sensors():
            for key, value in data._asdict().items():
                # From status
                if key == "percent":
                    continue
                self._database.write("battery", f"sensors_{key}", value)

    async def update_status(self) -> None:
        """Update Battery Status"""
        for key, value in self._battery.status().items():
            self._database.write("battery", camel_to_snake(key), value)

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_sensors(),
                self.update_status(),
            ]
        )



"""System Bridge: System"""
from __future__ import annotations

import os
import platform
import re
import socket
import uuid

from aiogithubapi import (
    GitHubAPI,
    GitHubConnectionException,
    GitHubException,
    GitHubRatelimitException,
    GitHubReleaseModel,
)
from pkg_resources import parse_version
from plyer import uniqueid
from psutil import boot_time, users
from psutil._common import suser
from systembridgeshared.base import Base
from systembridgeshared.database import Database

from systembridgebackend._version import __version__


class System(Base):
    """System"""

    def boot_time(self) -> float:
        """Get boot time"""
        return boot_time()

    def fqdn(self) -> str:
        """Get FQDN"""
        return socket.getfqdn()

    def hostname(self) -> str:
        """Get hostname"""
        return socket.gethostname()

    def ip_address_4(self) -> str:
        """Get IPv4 address"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(("8.8.8.8", 80))
        return sock.getsockname()[0]

    def mac_address(self) -> str:
        """Get MAC address"""
        # pylint: disable=consider-using-f-string
        return ":".join(re.findall("..", "%012x" % uuid.getnode()))

    def platform(self) -> str:
        """Get platform"""
        return platform.system()

    def platform_version(self) -> str:
        """Get platform version"""
        return platform.version()

    def uptime(self) -> float:
        """Get uptime"""
        return os.times()[0]

    def users(self) -> list[suser]:  # pylint: disable=unsubscriptable-object
        """Get users"""
        return users()

    def uuid(self) -> str:
        """Get UUID"""
        return uniqueid.id

    def version(self) -> str:
        """Get version"""
        return __version__.public()

    async def version_latest(self) -> GitHubReleaseModel | None:
        """Get latest version from GitHub"""
        self._logger.info("Get latest version from GitHub")
        try:
            async with GitHubAPI() as github:
                releases = await github.repos.releases.list("timmo001/system-bridge")
            return releases.data[0] if releases.data else None
        except (
            GitHubConnectionException,
            GitHubRatelimitException,
        ) as error:
            self._logger.error("Error getting data from GitHub: %s", error)
        except GitHubException as error:
            self._logger.exception(
                "Unexpected error getting data from GitHub: %s", error
            )
        return None

    def version_newer_available(
        self,
        database: Database,
    ) -> bool | None:
        """Check if newer version is available"""
        version = database.read_table_by_key("system", "version").to_dict(
            orient="records"
        )[0]["value"]
        latest_version = database.read_table_by_key("system", "version_latest").to_dict(
            orient="records"
        )[0]["value"]
        if version is not None and latest_version is not None:
            return parse_version(latest_version) > parse_version(version)
        return None



"""System Bridge: Module Base"""
from systembridgeshared.base import Base
from systembridgeshared.const import COLUMN_KEY, COLUMN_TIMESTAMP, COLUMN_VALUE
from systembridgeshared.database import Database


class ModuleUpdateBase(Base):
    """Module Base"""

    def __init__(
        self,
        database: Database,
        table: str,
    ):
        super().__init__()

        self._database = database
        self._database.create_table(
            table,
            [
                (COLUMN_KEY, "TEXT PRIMARY KEY"),
                (COLUMN_VALUE, "TEXT"),
                (COLUMN_TIMESTAMP, "DOUBLE"),
            ],
        )

    async def update_all_data(self) -> None:
        """Update data"""
        raise NotImplementedError()



"""System Bridge: Display"""
from __future__ import annotations

from systembridgeshared.base import Base
from systembridgeshared.const import (
    COLUMN_HARDWARE_NAME,
    COLUMN_HARDWARE_TYPE,
    COLUMN_KEY,
    COLUMN_NAME,
    COLUMN_VALUE,
)
from systembridgeshared.database import Database


class Display(Base):
    """Display"""

    def get_displays(
        self,
        database: Database,
    ) -> list[str]:
        """Get Displays"""
        displays = []
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "display" in item[COLUMN_HARDWARE_TYPE].lower()
                and item[COLUMN_HARDWARE_NAME] is not None
                and item[COLUMN_HARDWARE_NAME] not in displays
            ):
                displays.append(item[COLUMN_HARDWARE_NAME])
        return displays

    def pixel_clock(
        self,
        database: Database,
    ) -> float | None:
        """Display pixel clock"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "display" in item[COLUMN_HARDWARE_TYPE].lower()
                and "pixel" in item[COLUMN_NAME].lower()
                and "clock" in item[COLUMN_NAME].lower()
            ):
                self._logger.debug(
                    "Found display pixel clock: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def refresh_rate(
        self,
        database: Database,
    ) -> float | None:
        """Display refresh rate"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "display" in item[COLUMN_HARDWARE_TYPE].lower()
                and "refresh" in item[COLUMN_NAME].lower()
                and "rate" in item[COLUMN_NAME].lower()
            ):
                self._logger.debug(
                    "Found display refresh rate: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def resolution_horizontal(
        self,
        database: Database,
    ) -> int | None:
        """Display resolution horizontal"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "display" in item[COLUMN_HARDWARE_TYPE].lower()
                and "resolution" in item[COLUMN_NAME].lower()
                and "horizontal" in item[COLUMN_NAME].lower()
            ):
                self._logger.debug(
                    "Found display resolution horizontal: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None

    def resolution_vertical(
        self,
        database: Database,
    ) -> int | None:
        """Display resolution vertical"""
        for item in database.read_table("sensors").to_dict(orient="records"):
            if (
                item[COLUMN_HARDWARE_TYPE] is not None
                and "display" in item[COLUMN_HARDWARE_TYPE].lower()
                and "resolution" in item[COLUMN_NAME].lower()
                and "vertical" in item[COLUMN_NAME].lower()
            ):
                self._logger.debug(
                    "Found display resolution vertical: %s = %s",
                    item[COLUMN_KEY],
                    item[COLUMN_VALUE],
                )
                return item[COLUMN_VALUE]
        return None



"""System Bridge: Server Handler - Media"""
from __future__ import annotations

import mimetypes
import os

import aiofiles
from plyer import storagepath
from sanic.request import Request
from sanic.response import HTTPResponse, file, json

QUERY_BASE = "base"
QUERY_PATH = "path"
QUERY_FILENAME = "filename"

BASE_DIRECTORIES = {
    "documents": storagepath.get_documents_dir(),
    "downloads": storagepath.get_downloads_dir(),
    "home": storagepath.get_home_dir(),
    "music": storagepath.get_music_dir(),
    "pictures": storagepath.get_pictures_dir(),
    "videos": storagepath.get_videos_dir(),
}


def get_directories() -> list[dict]:
    """Get directories"""
    directories = []
    for key, value in BASE_DIRECTORIES.items():
        directories.append(
            {
                "key": key,
                "path": value,
            }
        )
    return directories


def get_files(
    base_path: str,
    path: str,
) -> list[dict]:
    """Get files from path"""
    files_info = []
    for filename in os.listdir(path):
        file_info = get_file(BASE_DIRECTORIES[base_path], os.path.join(path, filename))
        if file_info is not None:
            files_info.append(file_info)

    return files_info


def get_file(
    base_path: str,
    filepath: str,
) -> dict | None:
    """Get file from path"""
    try:
        stat = os.stat(filepath)
    except FileNotFoundError:
        return None

    mime_type = None
    if os.path.isfile(filepath):
        mime_type = mimetypes.guess_type(filepath)[0]

    return {
        "name": os.path.basename(filepath),
        "path": filepath.removeprefix(base_path)[1:],
        "fullpath": filepath,
        "size": stat.st_size,
        "last_accessed": stat.st_atime,
        "created": stat.st_ctime,
        "modified": stat.st_mtime,
        "is_directory": os.path.isdir(filepath),
        "is_file": os.path.isfile(filepath),
        "is_link": os.path.islink(filepath),
        "mime_type": mime_type,
    }


async def get_file_data(
    filepath: str,
) -> HTTPResponse:
    """Get file data"""
    return await file(filepath)


async def handler_media_directories(
    _: Request,
) -> HTTPResponse:
    """Handler for media directories"""
    return json(
        {
            "directories": get_directories(),
        }
    )


async def handler_media_files(
    request: Request,
) -> HTTPResponse:
    """Handler for media files"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )
    query_path = request.args.get(QUERY_PATH)
    path = (
        os.path.join(BASE_DIRECTORIES[query_base], query_path)
        if query_path
        else BASE_DIRECTORIES[query_base]
    )
    if not os.path.exists(path):
        return json(
            {"message": "Cannot find path", "path": path},
            status=404,
        )
    if not os.path.isdir(path):
        return json(
            {"message": "Path is not a directory", "path": path},
            status=400,
        )

    return json(
        {
            "files": get_files(query_base, path),
            "path": path,
        }
    )


async def handler_media_file(
    request: Request,
) -> HTTPResponse:
    """Handler for media file requests"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )
    if not (query_path := request.args.get(QUERY_PATH)):
        return json(
            {"message": "No path specified"},
            status=400,
        )
    path = os.path.join(BASE_DIRECTORIES[query_base], query_path)
    if not os.path.exists(path):
        return json(
            {"message": "Cannot find path", "path": path},
            status=404,
        )
    if not os.path.isfile(path):
        return json(
            {"message": "Path is not a file", "path": path},
            status=400,
        )

    return json(get_file(BASE_DIRECTORIES[query_base], path))


async def handler_media_file_data(
    request: Request,
) -> HTTPResponse:
    """Handler for media file requests"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )
    query_path = request.args.get(QUERY_PATH)
    path = os.path.join(BASE_DIRECTORIES[query_base], query_path)
    if not path:
        return json(
            {"message": "Cannot find path", "path": path},
            status=400,
        )
    if not os.path.exists(path):
        return json(
            {"message": "File does not exist", "path": path},
            status=404,
        )
    if not os.path.isfile(path):
        return json(
            {"message": "Path is not a file", "path": path},
            status=400,
        )

    return await get_file_data(path)


async def handler_media_file_write(
    request: Request,
) -> HTTPResponse:
    """Handler for media file write requests"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )
    if not (query_path := request.args.get(QUERY_PATH)):
        return json(
            {"message": "No path specified"},
            status=400,
        )
    if not (query_filename := request.args.get(QUERY_FILENAME)):
        return json(
            {"message": "No filename specified"},
            status=400,
        )
    path = os.path.join(BASE_DIRECTORIES[query_base], query_path)
    if not path:
        return json(
            {"message": "Cannot find path", "path": path},
            status=400,
        )
    if not request.body:
        return json(
            {"message": "No file specified"},
            status=400,
        )

    if not os.path.exists(path):
        os.makedirs(path)
    async with aiofiles.open(os.path.join(path, query_filename), "wb") as new_file:
        await new_file.write(request.body)
        await new_file.close()

    return json(
        {
            "message": "File uploaded",
            "path": path,
            "filename": query_filename,
        }
    )



"""System Bridge: Server Handler - Keyboard"""
from keyboard import press_and_release, write
from sanic.request import Request
from sanic.response import HTTPResponse, json


def keyboard_keypress(key: str):
    """Press a keyboard key"""
    press_and_release(key)


def keyboard_text(text: str):
    """Type text"""
    write(text)


async def handler_keyboard(
    request: Request,
) -> HTTPResponse:
    """Send a keyboard event."""
    if request.json is None:
        return json(
            {
                "mesage": "Missing JSON body",
            },
            status=400,
        )

    if "key" in request.json:
        try:
            keyboard_keypress(request.json["key"])
        except ValueError as err:
            return json(
                {
                    "message": err.args[0],
                },
                status=400,
            )

        return json(
            {
                "message": "Keypress sent",
                "key": request.json["key"],
            }
        )
    if "text" in request.json:
        keyboard_text(request.json["text"])
        return json(
            {
                "message": "Text sent",
                "text": request.json["text"],
            }
        )

    return json(
        {
            "message": "No key or text provided",
        },
        status=400,
    )



"""System Bridge: Server Handler - Open"""
import os
import subprocess
import sys
from webbrowser import open_new_tab

from sanic.request import Request
from sanic.response import HTTPResponse, json


def open_path(
    path: str,
) -> None:
    """Open a file."""
    if sys.platform == "win32":
        os.startfile(path)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, path])


def open_url(
    url: str,
) -> None:
    """Open a URL in the default browser."""
    open_new_tab(url)


async def handler_open(
    request: Request,
) -> HTTPResponse:
    """Open a file or a URL in the default browser."""
    if request.json is None:
        return json(
            {
                "mesage": "Missing JSON body",
            },
            status=400,
        )

    if "path" in request.json:
        open_path(request.json["path"])
        return json(
            {
                "message": f"Opening path: {request.json['path']}",
            }
        )
    if "url" in request.json:
        open_url(request.json["url"])
        return json(
            {
                "message": f"Opening URL: {request.json['url']}",
            }
        )

    return json(
        {
            "message": "No path or url provided",
        },
        status=400,
    )



"""System Bridge: Update CPU"""
import asyncio

from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.cpu import CPU


class CPUUpdate(ModuleUpdateBase):
    """CPU Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "cpu")
        self._cpu = CPU()

    async def update_count(self) -> None:
        """Update CPU count"""
        self._database.write("cpu", "count", self._cpu.count())

    async def update_frequency(self) -> None:
        """Update CPU frequency"""
        for key, value in self._cpu.freq()._asdict().items():
            self._database.write("cpu", f"frequency_{key}", value)

    async def update_frequency_per_cpu(self) -> None:
        """Update CPU frequency per CPU"""
        count = 0
        for data in [freq._asdict() for freq in self._cpu.freq_per_cpu()]:
            for key, value in data.items():
                self._database.write("cpu", f"frequency_{count}_{key}", value)
            count += 1

    async def update_load_average(self) -> None:
        """Update load average"""
        avg_tuple = self._cpu.load_average()
        result = sum([avg_tuple[0], avg_tuple[1], avg_tuple[2]]) / 3
        self._database.write("cpu", "load_average", result)

    async def update_stats(self) -> None:
        """Update stats"""
        for key, value in self._cpu.stats()._asdict().items():
            self._database.write("cpu", f"stats_{key}", value)

    async def update_temperature(self) -> None:
        """Update temperature"""
        self._database.write(
            "cpu", "temperature", self._cpu.temperature(self._database)
        )

    async def update_times(self) -> None:
        """Update times"""
        for key, value in self._cpu.times()._asdict().items():
            self._database.write("cpu", f"times_{key}", value)

    async def update_times_percent(self) -> None:
        """Update times percent"""
        for key, value in self._cpu.times_percent()._asdict().items():
            self._database.write("cpu", f"times_percent_{key}", value)

    async def update_times_per_cpu(self) -> None:
        """Update times per CPU"""
        count = 0
        for data in [freq._asdict() for freq in self._cpu.times_per_cpu()]:
            for key, value in data.items():
                self._database.write("cpu", f"times_per_cpu_{count}_{key}", value)
            count += 1

    async def update_times_per_cpu_percent(self) -> None:
        """Update times per CPU percent"""
        count = 0
        for data in [freq._asdict() for freq in self._cpu.times_per_cpu_percent()]:
            for key, value in data.items():
                self._database.write(
                    "cpu", f"times_per_cpu_percent_{count}_{key}", value
                )
            count += 1

    async def update_usage(self) -> None:
        """Update usage"""
        self._database.write("cpu", "usage", self._cpu.usage())

    async def update_usage_per_cpu(self) -> None:
        """Update usage per CPU"""
        count = 0
        for value in self._cpu.usage_per_cpu():
            self._database.write("cpu", f"usage_{count}", value)
            count += 1

    async def update_voltage(self) -> None:
        """Update voltage"""
        self._database.write("cpu", "voltage", self._cpu.voltage(self._database))

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_count(),
                self.update_frequency(),
                self.update_frequency_per_cpu(),
                self.update_load_average(),
                self.update_stats(),
                self.update_temperature(),
                self.update_times(),
                self.update_times_percent(),
                self.update_times_per_cpu(),
                self.update_times_per_cpu_percent(),
                self.update_usage(),
                self.update_usage_per_cpu(),
                self.update_voltage(),
            ]
        )



"""System Bridge: Update Sensors"""
import asyncio

from systembridgeshared.base import Base
from systembridgeshared.common import make_key
from systembridgeshared.const import (
    COLUMN_HARDWARE_NAME,
    COLUMN_HARDWARE_TYPE,
    COLUMN_KEY,
    COLUMN_NAME,
    COLUMN_TIMESTAMP,
    COLUMN_TYPE,
    COLUMN_VALUE,
)
from systembridgeshared.database import Database

from systembridgebackend.modules.sensors import Sensors


class SensorsUpdate(Base):
    """Sensors Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__()

        self._database = database
        self._database.create_table(
            "sensors",
            [
                (COLUMN_KEY, "TEXT PRIMARY KEY"),
                (COLUMN_TYPE, "TEXT"),
                (COLUMN_NAME, "TEXT"),
                (COLUMN_HARDWARE_TYPE, "TEXT"),
                (COLUMN_HARDWARE_NAME, "TEXT"),
                (COLUMN_VALUE, "TEXT"),
                (COLUMN_TIMESTAMP, "DOUBLE"),
            ],
        )
        self._sensors = Sensors()

    async def update_fans(self) -> None:
        """Update Fan Sensors"""
        if data := self._sensors.fans():
            for key, value in data.items():
                for item in value:
                    for subkey, subvalue in item._asdict().items():
                        self._database.write_sensor(
                            "sensors",
                            f"fans_{key}_{subkey}",
                            subkey,
                            subkey,
                            key,
                            key,
                            subvalue,
                        )

    async def update_temperatures(self) -> None:
        """Update Temperature Sensors"""
        if data := self._sensors.temperatures():
            for key, value in data.items():
                for item in value:
                    for subkey, subvalue in item._asdict().items():
                        self._database.write_sensor(
                            "sensors",
                            f"temperatures_{key}_{subkey}",
                            subkey,
                            subkey,
                            key,
                            key,
                            subvalue,
                        )

    async def update_windows_sensors(self) -> None:
        """Update Windows Sensors"""
        if not (data := self._sensors.windows_sensors()):
            return
        if "hardware" in data and data["hardware"] is not None:
            for hardware in data["hardware"]:
                key_hardware = (
                    make_key(f"_{hardware['name']}") if "name" in hardware else None
                )
                for sensor in hardware["sensors"] or []:
                    key_sensor_name = make_key(sensor["name"])
                    key_sensor_type = make_key(sensor["type"])

                    self._database.write_sensor(
                        "sensors",
                        f"windows_hardware{key_hardware}_{key_sensor_name}_{key_sensor_type}",
                        sensor["type"],
                        sensor["name"],
                        hardware["type"],
                        hardware["name"],
                        sensor["value"] if "value" in sensor else None,
                    )

        if "nvidia" in data and data["nvidia"] is not None:
            for sensor, value in data["nvidia"].items():
                if isinstance(value, list):
                    counter = 0
                    for hardware in value:
                        key_hardware = (
                            make_key(f"_{hardware['name']}")
                            if "name" in hardware
                            else None
                        )
                        type_hardware = (
                            hardware["type"] if "type" in hardware else "NVIDIA"
                        )
                        name_hardware = hardware["name"] if "name" in hardware else ""
                        if "DISPLAY" in name_hardware:
                            type_hardware = "Display"
                            name_hardware = (
                                f"Display {name_hardware.split('DISPLAY')[1]}"
                            )
                        for subkey, subvalue in hardware.items():
                            self._database.write_sensor(
                                "sensors",
                                f"windows_nvidia{key_hardware}_{sensor}_{counter}_{subkey}",
                                sensor,
                                subkey,
                                type_hardware,
                                name_hardware,
                                subvalue,
                            )
                        counter += 1
                else:
                    for subkey, subvalue in value.items():
                        self._database.write_sensor(
                            "sensors",
                            f"windows_nvidia_{sensor}_{subkey}",
                            sensor,
                            subkey,
                            value["type"] if "type" in value else "NVIDIA",
                            value["name"] if "name" in value else "NVIDIA",
                            subvalue,
                        )

    async def update_all_data(self) -> None:
        """Update data"""

        # Clear table in case of hardware changes since last run
        self._database.clear_table("sensors")

        await asyncio.gather(
            *[
                self.update_fans(),
                self.update_temperatures(),
                self.update_windows_sensors(),
            ]
        )



"""
Provides systembridgebackend version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update systembridgebackend` to change this file.

from incremental import Version

__version__ = Version("systembridgebackend", 3, 1, 5)
__all__ = ["__version__"]



"""System Bridge: Server"""
import asyncio
from collections.abc import Callable
from datetime import timedelta
import os
from os import walk
import sys

from sanic import Sanic
from sanic.request import Request
from sanic.response import HTTPResponse, json
from sanic_scheduler import SanicScheduler, task
from systembridgeshared.base import Base
from systembridgeshared.const import SECRET_API_KEY, SETTING_LOG_LEVEL, SETTING_PORT_API
from systembridgeshared.database import Database
from systembridgeshared.settings import Settings

from systembridgebackend.data import Data
from systembridgebackend.gui import GUIAttemptsExceededException, start_gui_threaded
from systembridgebackend.modules.listeners import Listeners
from systembridgebackend.server.auth import ApiKeyAuthentication
from systembridgebackend.server.keyboard import handler_keyboard
from systembridgebackend.server.mdns import MDNSAdvertisement
from systembridgebackend.server.media import (
    handler_media_directories,
    handler_media_file,
    handler_media_file_data,
    handler_media_file_write,
    handler_media_files,
)
from systembridgebackend.server.notification import handler_notification
from systembridgebackend.server.open import handler_open
from systembridgebackend.server.power import (
    handler_hibernate,
    handler_lock,
    handler_logout,
    handler_restart,
    handler_shutdown,
    handler_sleep,
)
from systembridgebackend.server.update import handler_update
from systembridgebackend.server.websocket import WebSocketHandler


class ApplicationExitException(BaseException):
    """Forces application to close."""


class Server(Base):
    """Server"""

    def __init__(
        self,
        database: Database,
        settings: Settings,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database
        self._settings = settings
        self._server = Sanic("SystemBridge")

        implemented_modules = []
        for _, dirs, _ in walk(os.path.join(os.path.dirname(__file__), "../modules")):
            implemented_modules = list(filter(lambda d: "__" not in d, dirs))
            break

        SanicScheduler(self._server, utc=True)
        self._listeners = Listeners(self._database, implemented_modules)
        self._data = Data(self._database, self._data_updated)

        auth = ApiKeyAuthentication(
            app=self._server,
            arg="apiKey",
            header="api-key",
            keys=[self._settings.get_secret(SECRET_API_KEY)],
        )

        mdns_advertisement = MDNSAdvertisement(self._settings)
        mdns_advertisement.advertise_server()

        @task(start=timedelta(seconds=2))
        async def _after_startup(_) -> None:
            """After startup"""
            if "--no-gui" not in sys.argv:
                try:
                    await start_gui_threaded(self._logger, self._settings)
                except GUIAttemptsExceededException:
                    self._logger.error("GUI could not be started. Exiting application")
                    await self._exit_application()

        @task(
            start=timedelta(seconds=10),
            period=timedelta(minutes=2),
        )
        async def _update_data(_) -> None:
            """Update data"""
            self._data.request_update_data()

        @task(
            start=timedelta(seconds=10),
            period=timedelta(seconds=30),
        )
        async def _update_frequent_data(_) -> None:
            """Update frequent data"""
            self._data.request_update_frequent_data()

        @auth.key_required
        async def _handler_data_all(
            _: Request,
            table: str,
        ) -> HTTPResponse:
            """Data handler all"""
            if table not in implemented_modules:
                return json({"message": f"Data module {table} not found"}, status=404)
            return json(self._database.table_data_to_ordered_dict(table))

        @auth.key_required
        async def _handler_data_by_key(
            _: Request,
            table: str,
            key: str,
        ) -> HTTPResponse:
            """Data handler by key"""
            if table not in implemented_modules:
                return json({"message": f"Data module {table} not found"}, status=404)

            data = self._database.read_table_by_key(table, key).to_dict(
                orient="records"
            )[0]
            return json(
                {
                    data["key"]: data["value"],
                    "last_updated": data["timestamp"],
                }
            )

        @auth.key_required
        async def _handler_generic(
            request: Request,
            function: Callable,
        ) -> HTTPResponse:
            """Generic handler"""
            return await function(request)

        async def _handler_websocket(
            _: Request,
            socket,
        ) -> None:
            """WebSocket handler"""
            websocket = WebSocketHandler(
                self._database,
                self._settings,
                self._listeners,
                implemented_modules,
                socket,
                self._callback_exit_application,
            )
            await websocket.handler()

        self._server.add_route(
            _handler_data_all,
            "/api/data/<table:str>",
            methods=["GET"],
        )
        self._server.add_route(
            _handler_data_by_key,
            "/api/data/<table:str>/<key:str>",
            methods=["GET"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_keyboard),
            "/api/keyboard",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_media_directories),
            "/api/media",
            methods=["GET"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_media_files),
            "/api/media/files",
            methods=["GET"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_media_file),
            "/api/media/file",
            methods=["GET"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_media_file_data),
            "/api/media/file/data",
            methods=["GET"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_media_file_write),
            "/api/media/file/write",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_notification),
            "/api/notification",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_open),
            "/api/open",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_sleep),
            "/api/power/sleep",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_hibernate),
            "/api/power/hibernate",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_restart),
            "/api/power/restart",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_shutdown),
            "/api/power/shutdown",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_lock),
            "/api/power/lock",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_logout),
            "/api/power/logout",
            methods=["POST"],
        )
        self._server.add_route(
            lambda r: _handler_generic(r, handler_update),
            "/api/update",
            methods=["POST"],
        )

        if "--no-frontend" not in sys.argv:
            try:
                # pylint: disable=import-error, import-outside-toplevel
                from systembridgefrontend import get_frontend_path

                frontend_path = get_frontend_path()
                self._logger.info("Serving frontend from: %s", frontend_path)
                self._server.static(
                    "/",
                    frontend_path,
                    strict_slashes=False,
                    content_type="text/html",
                )
            except (ImportError, ModuleNotFoundError) as error:
                self._logger.error("Frontend not found: %s", error)

        self._server.add_websocket_route(
            _handler_websocket,
            "/api/websocket",
        )

    def _callback_exit_application(self) -> None:
        """Callback to exit application"""
        asyncio.create_task(self._exit_application())

    async def _data_updated(
        self,
        module: str,
    ) -> None:
        """Data updated"""
        await self._listeners.refresh_data_by_module(module)

    async def _exit_application(self) -> None:
        """Exit application"""
        self._logger.info("Exiting application")
        self.stop_server()

    def start_server(self) -> None:
        """Start Server"""
        if (port := self._settings.get(SETTING_PORT_API)) is None:
            raise ValueError("Port not set")
        self._logger.info("Starting server on port: %s", port)
        self._server.run(
            host="0.0.0.0",
            port=int(port),
            access_log=False,
            debug=self._settings.get(SETTING_LOG_LEVEL) == "DEBUG",
            motd=False,
        )

    def stop_server(self) -> None:
        """Stop Server"""
        self._logger.info("Stopping server")
        loop = self._server.loop
        self._logger.info("Cancel any pending tasks")
        for pending_task in asyncio.all_tasks():
            pending_task.cancel()
        self._logger.info("Stop the event loop")
        loop.stop()
        # self._server.enable_websocket(False)
        self._listeners.remove_all_listeners()
        self._server.stop()



"""System Bridge: Autostart Windows"""
import platform
import sys


def autostart_windows_disable():
    """Disable autostart for Windows"""
    if platform.system() != "Windows":
        return

    # pylint: disable=import-error, import-outside-toplevel
    from winreg import HKEY_CURRENT_USER, KEY_ALL_ACCESS, CloseKey, DeleteValue, OpenKey

    try:
        key = OpenKey(
            HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Run",
            reserved=0,
            access=KEY_ALL_ACCESS,
        )
        if key is None:
            return
        DeleteValue(key, "systembridgebackend")
        CloseKey(key)
    except OSError:
        pass


def autostart_windows_enable():
    """Enable autostart for Windows"""
    if platform.system() != "Windows":
        return

    # pylint: disable=import-error, import-outside-toplevel
    from winreg import (
        HKEY_CURRENT_USER,
        KEY_ALL_ACCESS,
        REG_SZ,
        CloseKey,
        OpenKey,
        SetValueEx,
    )

    key = OpenKey(
        HKEY_CURRENT_USER,
        r"Software\Microsoft\Windows\CurrentVersion\Run",
        reserved=0,
        access=KEY_ALL_ACCESS,
    )
    SetValueEx(
        key,
        "systembridgebackend",
        0,
        REG_SZ,
        f'"{sys.executable}" -m systembridgebackend --silent',
    )
    CloseKey(key)



"""System Bridge: Shortcut Windows"""
import os
import platform
import sys


def create_windows_shortcuts():
    """Create Windows shortcuts"""
    if platform.system() != "Windows":
        return

    # pylint: disable=import-error, import-outside-toplevel
    from winreg import HKEY_CURRENT_USER, KEY_READ, CloseKey, OpenKey, QueryValueEx

    from win32com.client import Dispatch

    registry_key = OpenKey(
        HKEY_CURRENT_USER,
        r"Software\Microsoft\Windows\CurrentVersion\Explorer\User Shell Folders",
        reserved=0,
        access=KEY_READ,
    )
    registry_value, _ = QueryValueEx(registry_key, "Programs")
    CloseKey(registry_key)

    directory = os.path.abspath(
        os.path.join(
            os.path.expandvars(os.path.normpath(registry_value)),
            "systembridge",
        )
    )
    os.makedirs(directory, exist_ok=True)

    link_path = os.path.join(directory, "System Bridge.lnk")
    shell = Dispatch("WScript.Shell")
    shortcut = shell.CreateShortCut(link_path)
    shortcut.Arguments = "-m systembridgebackend --silent"
    shortcut.Description = "System Bridge"
    shortcut.IconLocation = os.path.join(os.path.dirname(__file__), "../icon.ico")
    shortcut.Targetpath = sys.executable
    shortcut.WorkingDirectory = sys.prefix
    shortcut.save()



"""System Bridge: Update Bridge"""
import time

from systembridgeshared.base import Base
from systembridgeshared.const import COLUMN_TIMESTAMP
from systembridgeshared.database import Database
from zeroconf import ServiceStateChange, Zeroconf

from systembridgebackend.modules.bridge import Bridge

COLUMN_NAME = "name"
COLUMN_ADDRESS = "address"
COLUMN_FQDN = "fqdn"
COLUMN_HOST = "host"
COLUMN_IP = "ip"
COLUMN_MAC = "mac"
COLUMN_PORT = "port"
COLUMN_UUID = "uuid"
COLUMN_VERSION = "version"
COLUMN_WEBSOCKET_ADDRESS = "websocket_address"
COLUMN_WEBSOCKET_PORT = "websocket_port"
COLUMN_ACTIVE = "active"
COLUMN_LAST_ACTIVE_TIMESTAMP = "last_active_timestamp"


class BridgeUpdate(Base):
    """Bridge Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__()

        self._database = database
        self._database.create_table(
            "bridge",
            [
                (COLUMN_UUID, "TEXT PRIMARY KEY"),
                (COLUMN_NAME, "TEXT"),
                (COLUMN_ADDRESS, "TEXT"),
                (COLUMN_FQDN, "TEXT"),
                (COLUMN_HOST, "TEXT"),
                (COLUMN_IP, "TEXT"),
                (COLUMN_MAC, "TEXT"),
                (COLUMN_PORT, "TEXT"),
                (COLUMN_VERSION, "TEXT"),
                (COLUMN_WEBSOCKET_ADDRESS, "TEXT"),
                (COLUMN_WEBSOCKET_PORT, "TEXT"),
                (COLUMN_ACTIVE, "INTEGER"),
                (COLUMN_LAST_ACTIVE_TIMESTAMP, "INTEGER"),
                (COLUMN_TIMESTAMP, "DOUBLE"),
            ],
        )
        self._bridge = Bridge(self.service_changed)

    def service_changed(
        self,
        zeroconf: Zeroconf,
        service_type: str,
        name: str,
        state_change: ServiceStateChange,
    ) -> None:
        """Service changed"""
        self._logger.info(
            "service_changed: type=%s name=%s state_change=%s",
            service_type,
            name,
            state_change,
        )
        service_info = zeroconf.get_service_info(service_type, name)
        self._logger.info(
            "Service %s %s: %s - %s", name, state_change, service_type, service_info
        )
        if service_info and service_info.properties:
            if state_change in (ServiceStateChange.Added, ServiceStateChange.Updated):
                self.write_bridge(name, service_info.properties)
            elif state_change == ServiceStateChange.Removed:
                self.deactivate_bridge(name, service_info.properties)

    def deactivate_bridge(
        self,
        name: str,
        info: dict,
    ) -> None:
        """Deactivate bridge"""
        self._logger.info("Deactivate bridge %s", name)
        self._database.execute_sql(
            f"""UPDATE bridge SET {COLUMN_ACTIVE} = 0, {COLUMN_TIMESTAMP} = {time.time()}
             WHERE {COLUMN_UUID}={info[b"uuid"].decode("utf-8")}""".replace(
                "\n", ""
            ).replace(
                "            ", ""
            ),
        )

    def write_bridge(
        self,
        name: str,
        info: dict,
    ) -> None:
        """Write bridge"""
        values = {
            COLUMN_UUID: info[b"uuid"].decode("utf-8"),
            COLUMN_NAME: name,
            COLUMN_ADDRESS: info[b"address"].decode("utf-8"),
            COLUMN_FQDN: info[b"fqdn"].decode("utf-8"),
            COLUMN_HOST: info[b"host"].decode("utf-8"),
            COLUMN_IP: info[b"ip"].decode("utf-8"),
            COLUMN_MAC: info[b"mac"].decode("utf-8"),
            COLUMN_PORT: info[b"port"].decode("utf-8"),
            COLUMN_VERSION: info[b"version"].decode("utf-8"),
            COLUMN_WEBSOCKET_ADDRESS: info[b"websocketAddress"].decode("utf-8"),
            COLUMN_WEBSOCKET_PORT: info[b"wsPort"].decode("utf-8"),
            COLUMN_ACTIVE: 1,
            COLUMN_LAST_ACTIVE_TIMESTAMP: time.time(),
            COLUMN_TIMESTAMP: time.time(),
        }

        self._database.execute_sql(
            f"""INSERT INTO bridge ({COLUMN_UUID}, {COLUMN_NAME}, {COLUMN_ADDRESS}, {COLUMN_FQDN},
             {COLUMN_HOST}, {COLUMN_IP}, {COLUMN_MAC}, {COLUMN_PORT}, {COLUMN_VERSION},
             {COLUMN_WEBSOCKET_ADDRESS}, {COLUMN_WEBSOCKET_PORT}, {COLUMN_ACTIVE},
             {COLUMN_LAST_ACTIVE_TIMESTAMP}, {COLUMN_TIMESTAMP})
             VALUES("{values[COLUMN_UUID]}", "{values[COLUMN_NAME]}", "{values[COLUMN_ADDRESS]}",
             "{values[COLUMN_FQDN]}", "{values[COLUMN_HOST]}", "{values[COLUMN_IP]}",
             "{values[COLUMN_MAC]}", "{values[COLUMN_PORT]}", "{values[COLUMN_VERSION]}",
             "{values[COLUMN_WEBSOCKET_ADDRESS]}", "{values[COLUMN_WEBSOCKET_PORT]}",
             {values[COLUMN_ACTIVE]}, {values[COLUMN_LAST_ACTIVE_TIMESTAMP]},
             {values[COLUMN_TIMESTAMP]})
             ON CONFLICT({COLUMN_UUID}) DO
             UPDATE SET {COLUMN_ADDRESS} = "{values[COLUMN_ADDRESS]}",
             {COLUMN_FQDN} = "{values[COLUMN_FQDN]}", {COLUMN_HOST} = "{values[COLUMN_HOST]}",
             {COLUMN_IP} = "{values[COLUMN_IP]}", {COLUMN_MAC} = "{values[COLUMN_MAC]}",
             {COLUMN_PORT} = "{values[COLUMN_PORT]}",
             {COLUMN_VERSION} = "{values[COLUMN_VERSION]}",
             {COLUMN_WEBSOCKET_ADDRESS} = "{values[COLUMN_WEBSOCKET_ADDRESS]}",
             {COLUMN_WEBSOCKET_PORT} = "{values[COLUMN_WEBSOCKET_PORT]}",
             {COLUMN_ACTIVE} = {values[COLUMN_ACTIVE]},
             {COLUMN_LAST_ACTIVE_TIMESTAMP} = {values[COLUMN_LAST_ACTIVE_TIMESTAMP]},
             {COLUMN_TIMESTAMP} = {values[COLUMN_TIMESTAMP]}
             WHERE {COLUMN_UUID} = "{values[COLUMN_UUID]}"
            """.replace(
                "\n", ""
            ).replace(
                "            ", ""
            )
        )



"""System Bridge: Server Handler - Notification"""
import os

from plyer import notification
from sanic.request import Request
from sanic.response import HTTPResponse, json


async def handler_notification(
    request: Request,
) -> HTTPResponse:
    """Send a notification."""
    if request.json is None:
        return json(
            {
                "mesage": "Missing JSON body",
            },
            status=400,
        )

    if "message" not in request.json:
        return json(
            {
                "message": "No message provided",
            },
            status=400,
        )

    notification.notify(
        title=request.json["title"] if "title" in request.json else "",
        message=request.json["message"],
        app_name=request.json["app_name"]
        if "app_name" in request.json
        else "System Bridge",
        app_icon=request.json["icon"]
        if "icon" in request.json
        else "./resources/system-bridge-circle.ico"
        if os.name == "nt"
        else "./resources/system-bridge-circle.png",
        timeout=request.json["timeout"] if "timeout" in request.json else 10,
    )

    return json(
        {
            "message": "Notification sent",
        }
    )



"""System Bridge: Main"""
import logging
import platform
import sys

from systembridgeshared.base import Base
from systembridgeshared.const import SETTING_AUTOSTART, SETTING_LOG_LEVEL
from systembridgeshared.database import Database
from systembridgeshared.logger import setup_logger
from systembridgeshared.settings import Settings

from systembridgebackend.autostart import autostart_disable, autostart_enable
from systembridgebackend.modules.system import System
from systembridgebackend.server import Server
from systembridgebackend.shortcut import create_shortcuts


class Main(Base):
    """Main"""

    def __init__(self) -> None:
        """Initialize"""
        super().__init__()
        if "--init" in sys.argv:
            self._logger.info("Initialized application. Exiting now.")
            sys.exit(0)

        self._logger.info("System Bridge %s: Startup", System().version())

        if "--cli" not in sys.argv:
            autostart = settings.get(SETTING_AUTOSTART)
            self._logger.info("Autostart enabled: %s", autostart)
            if autostart:
                autostart_enable()
            else:
                autostart_disable()

            create_shortcuts()

        if platform.system() == "Windows" and "--silent" in sys.argv:
            self._logger.info("Hide console")
            # pylint: disable=import-error, import-outside-toplevel
            from win32con import SW_HIDE
            from win32gui import GetForegroundWindow, ShowWindow

            ShowWindow(GetForegroundWindow(), SW_HIDE)  # type: ignore

        self._server = Server(database, settings)

        # Start the server
        self._server.start_server()


if __name__ == "__main__":

    database = Database()
    settings = Settings(database)

    LOG_LEVEL = str(settings.get(SETTING_LOG_LEVEL))
    setup_logger(LOG_LEVEL, "system-bridge")
    logging.getLogger("zeroconf").setLevel(logging.ERROR)

    Main()
